@@ -4,7 +4,8 @@ import numpy as np
 from numpy.random import seed
 from skimage.transform import resize
 from scipy.ndimage.filters import gaussian_filter
-from tqdm.auto import tqdm
+from tqdm.auto import tqdm, trange
+from tqdm.contrib import tenumerate
 import functools
 import requests
 import re
@@ -21,10 +22,11 @@ __all__ = [
     "volshow", "get_files", "get_mmr_fromfile",
     # useful utils
     "get_file", "load_file", "gunzip_array", "ellipsoid", "add_lesions",
+    "get_label_probabilities",
     # nothing to do with BrainWeb but still useful
-    "register",
+    "trim_zeros_ROI", "register",
     # intensities
-    "FDG", "Amyloid", "T1", "T2",
+    "Act", "FDG", "Amyloid", "T1", "T2", "Mu",
     # scanner params
     "Res", "Shape",
     # probably not needed
@@ -32,12 +34,13 @@ __all__ = [
 
 LINKS = "04 05 06 18 20 38 41 42 43 44 45 46 47 48 49 50 51 52 53 54"
 LINKS = [
-    'http://brainweb.bic.mni.mcgill.ca/cgi/brainweb1' +
-    '?do_download_alias=subject' + i + '_crisp&format_value=raw_short' +
+    'http://brainweb.bic.mni.mcgill.ca/cgi/brainweb1'
+    '?do_download_alias=subject' + i + '_crisp&format_value=raw_short'
     '&zip_value=gnuzip&download_for_real=%5BStart+download%21%5D'
     for i in LINKS.split()]
 RE_SUBJ = re.compile('.*(subject)([0-9]+).*')
 LINKS = dict((RE_SUBJ.sub(r'\1_\2.bin.gz', i), i) for i in LINKS)
+log = logging.getLogger(__name__)
 
 
 class Act(object):
@@ -45,6 +48,9 @@ class Act(object):
   background, csf, greyMatter, whiteMatter, fat, muscle, skin, skull, vessels,\
       aroundFat, dura, marrow\
       = [i << 4 for i in range(12)]
+  all_labels = [
+      'background', 'csf', 'greyMatter', 'whiteMatter', 'fat', 'muscle',
+      'skin', 'skull', 'vessels', 'aroundFat', 'dura', 'marrow']
   bone = skull | marrow | dura
 
   @classmethod
@@ -134,7 +140,7 @@ class Shape(object):
   brainweb = mMR * Res.mMR / Res.brainweb
 
 
-def get_file(fname, origin, cache_dir=None):
+def get_file(fname, origin, cache_dir=None, chunk_size=None):
     """
     Downloads a file from a URL if it not already in the cache.
     By default the file at the url `origin` is downloaded to the
@@ -152,8 +158,6 @@ def get_file(fname, origin, cache_dir=None):
         defaults to `~/.brainweb`.
     @return  : Path to the downloaded file
     """
-    log = logging.getLogger(__name__)
-
     if cache_dir is None:
         cache_dir = path.join('~', '.brainweb')
     cache_dir = path.expanduser(cache_dir)
@@ -173,11 +177,12 @@ def get_file(fname, origin, cache_dir=None):
         log.debug("Downloading %s from %s" % (fpath, origin))
         try:
             d = requests.get(origin, stream=True)
-            with tqdm(total=d.headers.get('Content-length', None), desc=fname,
+            with tqdm(total=float(d.headers.get('Content-length') or 0),
+                      desc=fname,
                       unit="B", unit_scale=True, unit_divisor=1024,
                       leave=False) as fprog:
                 with open(fpath, 'wb') as fo:
-                    for chunk in d.iter_content(chunk_size=None):
+                    for chunk in d.iter_content(chunk_size=chunk_size):
                         fo.write(chunk)
                         fprog.update(len(chunk))
                 fprog.total = fprog.n
@@ -218,17 +223,22 @@ def get_files(cache_dir=None, progress=True):
 def get_mmr(cache_file, raw_data,
             petNoise=1.0, t1Noise=0.75, t2Noise=0.75,
             petSigma=1.0, t1Sigma=1.0, t2Sigma=1.0,
-            PetClass=FDG):
+            outres="mMR", PetClass=FDG):
     """
     Return contents of specified `*.npz` file,
     creating it from BrainWeb `raw_data` 3darray if it doesn't exist.
+
+    @param outres: attribute to use from Res/Class [default: "mMR"]
     """
     if not path.exists(cache_file):
-        pet, uMap, t1, t2 = toPetMmr(raw_data, PetClass=PetClass)
+        pet, uMap, t1, t2 = toPetMmr(
+            raw_data, outres=outres, PetClass=PetClass)
         pet = noise(pet, petNoise, sigma=petSigma)[:, ::-1]
         t1 = noise(t1, t1Noise, sigma=t1Sigma)[:, ::-1]
         t2 = noise(t2, t2Noise, sigma=t2Sigma)[:, ::-1]
         uMap = uMap[:, ::-1]
+        out_res = getattr(Res, outres)
+
         np.savez_compressed(
             cache_file,
             PET=pet, uMap=uMap, T1=t1, T2=t2,
@@ -237,28 +247,66 @@ def get_mmr(cache_file, raw_data,
             t2Noise=np.float32(t2Noise),
             petSigma=np.float32(petSigma),
             t1Sigma=np.float32(t1Sigma),
-            t2Sigma=np.float32(t2Sigma))
+            t2Sigma=np.float32(t2Sigma),
+            res=np.float32(out_res))
 
-    return np.load(cache_file, allow_pickle=True)
+    cached = np.load(cache_file, allow_pickle=True)
+    if 'res' not in cached:
+        log.info("%s:converting old data format", cache_file)
+        data = dict(cached)
+        data['res'] = getattr(Res, 'mMR')
+        np.savez_compressed(cache_file, **data)
+        cached = np.load(cache_file, allow_pickle=True)
+
+    return cached
 
 
 def get_mmr_fromfile(brainweb_file,
                      petNoise=1.0, t1Noise=0.75, t2Noise=0.75,
                      petSigma=1.0, t1Sigma=1.0, t2Sigma=1.0,
-                     PetClass=FDG):
+                     outres="mMR", PetClass=FDG):
     """
-    mMR resolution ground truths from a cached `np.load`able file generated
+    ground truths from a cached `np.load`able file generated
     from `brainweb_file`.
+
+    @param outres: attribute to use from Res/Class [default: "mMR"]
     """
     dat = load_file(brainweb_file)  # read raw data
     return get_mmr(
         brainweb_file.replace(
-            '.bin.gz', '.npz' if PetClass == FDG else
-            '.{}.npz'.format(PetClass.__name__)),
+            '.bin.gz', ('.%s.%s.npz' % (PetClass.__name__, outres)).replace(
+                '.FDG.mMR', '')),
         dat,
-        petNoise=1.0, t1Noise=0.75, t2Noise=0.75,
-        petSigma=1.0, t1Sigma=1.0, t2Sigma=1.0,
-        PetClass=PetClass)
+        petNoise=petNoise, t1Noise=t1Noise, t2Noise=t2Noise,
+        petSigma=petSigma, t1Sigma=t1Sigma, t2Sigma=t2Sigma,
+        outres=outres, PetClass=PetClass)
+
+
+def get_label_probabilities(brainweb_file, labels=None, outres="mMR",
+                            progress=True, dtype=np.float32):
+    """
+    @param labels  : list of strings, [default: Act.all_labels]
+    @return out  : 4D array of masks resampled as per `outres` (useful for PVC)
+    """
+    out_shape = getattr(Shape, outres)
+    raw_data = load_file(brainweb_file)
+    if labels is None:
+        labels = Act.all_labels
+    if set(labels).difference(Act.all_labels):
+        raise KeyError(
+            "labels (%s) must be in Act.all_labels (%s)" % (
+                ", ".join(labels), ", ".join(Act.all_labels)))
+
+    num_classes = len(labels)
+    res = np.zeros((num_classes,) + tuple(out_shape), dtype=dtype)
+    for i, attr in tenumerate(labels, unit="label", desc="BrainWeb labels",
+                              disable=not progress):
+        class MAct(Act):
+          attrs = [attr]
+        setattr(MAct, attr, 1)
+        res[i] = toPetMmr(raw_data, outres=outres, modes=[MAct])[0][:, ::-1]
+
+    return res
 
 
 def volshow(vol,
@@ -284,9 +332,9 @@ def volshow(vol,
     @param tight_layout  : number of times to run `tight_layout(0, 0, 0)`
       [default: 1]
     """
-    import matplotlib.pyplot as plt
+    from IPython.display import display
     import ipywidgets as ipyw
-    log = logging.getLogger(__name__)
+    import matplotlib.pyplot as plt
 
     if hasattr(vol, "keys") and hasattr(vol, "values"):
         if titles is not None:
@@ -300,7 +348,7 @@ def volshow(vol,
     else:
         for v in vol:
             if v.ndim not in [3, 4]:
-                raise IndexError("Input should be (one or a list of)" +
+                raise IndexError("Input should be (one or a list of)"
                                  " 3D and/or 4D array(s)")
 
     if cmaps is None:
@@ -332,7 +380,13 @@ def volshow(vol,
         nrows = ncols = 2
 
     zSize = min(len(i) for i in vol)
-    fig = plt.figure(figsize=figsize, frameon=frameon)
+
+    # matplotlib>=3.3.2 figure updating needs its own output area
+    # https://github.com/matplotlib/matplotlib/issues/18638
+    out = ipyw.Output()
+    display(out)
+    with out:
+        fig = plt.figure(figsize=figsize, frameon=frameon)
 
     @ipyw.interact(z=ipyw.IntSlider(zSize // 2, 0, zSize - 1, 1))
     def plot_slice(z):
@@ -384,7 +438,6 @@ def noise(im, n, warn_zero=True, sigma=1):
   @param sigma  : float, smoothing of noise component
   @return[out] im  : array like im with +-n *100%im noise added
   """
-  log = logging.getLogger(__name__)
   if n < 0:
     raise ValueError("Noise must be positive")
   elif n == 0:
@@ -398,43 +451,16 @@ def noise(im, n, warn_zero=True, sigma=1):
 def toPetMmr(im, pad=True, dtype=np.float32, outres="mMR", modes=None,
              PetClass=FDG):
   """
+  @param outres: attribute to use from `Res` & `Shape` classes [default: "mMR"]
   @param modes  : [default: [PetClass, Mu, T1, T2]]
-  @return out  : list of `modes`, each shape [127, 344, 344]
+  @param PetClass  : [default: FDG]. Ignored if `modes` is explicitly specified
+  @return out  : 4D array with same length as `modes`
   """
-  log = logging.getLogger(__name__)
-
+  if modes is None:
+    modes = [PetClass, Mu, T1, T2]
   out_res = getattr(Res, outres)
   out_shape = getattr(Shape, outres)
 
-  if modes is None:
-    modes = [PetClass, Mu, T1, T2]
-
-  # PET
-  # res = np.zeros(im.shape, dtype=dtype)
-  res = np.zeros_like(im)
-  for attr in PetClass.attrs:
-    log.debug("PET:%s:%d" % (attr, getattr(PetClass, attr)))
-    res[Act.indices(im, attr)] = getattr(PetClass, attr)
-
-  # uMap
-  uMap = np.zeros(im.shape, dtype=dtype)
-  for attr in Mu.attrs:
-    log.debug("uMap:%s:%d" % (attr, getattr(Mu, attr)))
-    uMap[Act.indices(im, attr)] = getattr(Mu, attr)
-
-  # MR
-  # t1 = np.zeros(im.shape, dtype=dtype)
-  t1 = np.zeros_like(im)
-  for attr in T1.attrs:
-    log.debug("T1:%s:%d" % (attr, getattr(T1, attr)))
-    t1[Act.indices(im, attr)] = getattr(T1, attr)
-  # t2 = np.zeros(im.shape, dtype=dtype)
-  t2 = np.zeros_like(im)
-  for attr in T2.attrs:
-    log.debug("T2:%s:%d" % (attr, getattr(T2, attr)))
-    t2[Act.indices(im, attr)] = getattr(T2, attr)
-
-  # resize
   new_shape = np.rint(np.asarray(im.shape) * Res.brainweb / out_res)
   padLR, padR = divmod((np.array(out_shape) - new_shape), 2)
 
@@ -451,7 +477,15 @@ def toPetMmr(im, pad=True, dtype=np.float32, outres="mMR", modes=None,
                    mode="constant")
     return arr
 
-  return [resizeToMmr(i) for i in [res, uMap, t1, t2]]
+  res = np.zeros((len(modes),) + tuple(out_shape))
+  for i, MClass in enumerate(modes):
+    arr = np.zeros_like(im, dtype=dtype)  # dtype only needed for uMap?
+    for attr in MClass.attrs:
+      log.debug("%s:%s:%d" % (MClass.__name__, attr, getattr(MClass, attr)))
+      arr[Act.indices(im, attr)] = getattr(MClass, attr)
+    res[i] = resizeToMmr(arr)
+
+  return res
 
 
 def ellipsoid(out_shape, radii, position, dtype=np.float32):
@@ -534,6 +568,31 @@ def matify(mat, dtype=np.float32, transpose=None):
   return mat.transpose(transpose).astype(dtype)
 
 
+def trim_zeros_ROI(arr, trim='fb', progress=True):
+  """Returns a ROI corresponding to `numpy.trim_zeros`. Works on ndarrays."""
+  trim = trim.upper()
+  res = []
+  for i in trange(arr.ndim, desc="Trimming ROI",
+                  disable=not progress, leave=False):
+    filt = np.swapaxes(arr, i, 0) if i else arr
+    first = 0
+    if 'F' in trim:
+      for i in filt:
+        if i.any():
+          break
+        else:
+          first += 1
+    last = len(filt)
+    if 'B' in trim:
+      for i in filt[::-1]:
+        if i.any():
+          break
+        else:
+          last -= 1
+    res.append(slice(first, last))
+  return tuple(res)
+
+
 def register(src, target=None, ROI=None, target_shape=Shape.mMR,
              src_resolution=Res.MR, target_resolution=Res.mMR,
              method="CoM",
@@ -559,7 +618,6 @@ def register(src, target=None, ROI=None, target_shape=Shape.mMR,
       [default: "CoM"]  : centre of mass.
     """
     from dipy.align.imaffine import AffineMap, transform_centers_of_mass
-    log = logging.getLogger(__name__)
 
     assert src.ndim == 3
     if target is not None:
